import {expect} from 'chai'
import {diffJson} from 'diff'

import {FakeGenerator} from './FakeGenerator'

let _instance
class FetchFetcherFakerManager {
  _fakeRequests
  _methods = ['get', 'patch', 'post']

  static create(FetchFetcher) {
    if (_instance) return _instance
    _instance = new FetchFetcherFakerManager(FetchFetcher)
    return _instance
  }

  constructor(FetchFetcher) {
    this.FetchFetcher = FetchFetcher
    this._originalsFetcher = this._methods.map(
      method => this.FetchFetcher.prototype[method]
    )
    this._fakeRequests = {}
    this._methods.forEach(method => {
      this.FetchFetcher.prototype[method] = this._request(method)
    })
  }

  addMock({url, method, error, mock, force}) {
    const key = method.toUpperCase() + '::' + url
    if (!force && this._fakeRequests[key])
      throw new Error(
        `[FetchFetcherFakerManager#addMock] forbidden overwrite the request ${key}`
      )

    this._fakeRequests[key] = Promise.resolve([error, mock])
  }

  validate({url, method}) {
    const key = method.toUpperCase() + '::' + url
    if (this._fakeRequests[key])
      throw new Error(
        `[FetchFetcherFakerManager#validate] request ${key} don't consume`
      )
  }

  restore() {
    if (Object.keys(this._fakeRequests).length === 0) {
      this._methods.forEach(
        (method, index) =>
          (this.FetchFetcher.prototype[method] = this._originalsFetcher[index])
      )
    } else {
      console.warn(`[FetchFetcherFakerManager#restore]
        Unabled restore the FetchFetcher because there are request w/out been consume
        - Dont match any mock:
        ${Object.keys(this._fakeRequests).join('\n\t\t')}
      `)
    }
  }

  _request(method) {
    const self = this
    return function (options) {
      const requestKey = method.toUpperCase() + '::' + options.url
      // this === FetchFetcher instance
      if (self._fakeRequests[requestKey]) {
        const response = self._fakeRequests[requestKey]
        delete self._fakeRequests[requestKey]
        return response
      } else {
        console.warn(`[FetchFetcherFakerManager#_request]
        - Request make ${requestKey}
        - Dont match any mock:
        ${Object.keys(self._fakeRequests).join('\n\t\t')}
      `)
      }

      const {url, ...innerOptions} = options
      return this._request(url, {...innerOptions, method})
    }
  }
}

export class MotherObject {}

export class ModelMotherObject extends MotherObject {
  _model

  withoutError({error} = {}) {
    expect(error).to.be.eql(null)
    return this
  }

  withError({error, klass} = {}) {
    expect(error).instanceof(klass)
    return this
  }

  isNull({json} = {}) {
    expect(json).to.be.eql(null)
  }

  equals({json} = {}) {
    try {
      expect(this.jsonModel()).to.be.eql(json)
    } catch (err) {
      // pretty print this._model.toJSON() / json
      const diff = diffJson(this.jsonModel(), json)
      diff.forEach(part => {
        console.log(
          `%c ${part.value}`,
          `color: ${part.added ? 'green' : part.removed ? 'red' : 'grey'}`
        )
      })
      throw err
    }
    return this
  }

  jsonModel() {
    if (!this._model)
      throw new Error(
        `[ModelMotherObject#jsonModel] model is undefined model(${this._model})`
      )
    return this._model.toJSON()
  }
}

export const DataSourceMotherObjectFactory = DataSource =>
  class DataSourceMotherObject extends MotherObject {
    _fillDataSource
    _originalData
    _faker
    _mock

    constructor() {
      super()
      this._faker = FakeGenerator.create()
      this._originalData = DataSource.prototype.data
      DataSource.prototype.data = this._response.bind(this)
    }

    success() {
      if (this._fillDataSource)
        throw new Error(
          '[DataSourceMotherObject#success] Call success w/out been consume a previous call'
        )

      this._mock = this.generate(this._faker)
      this._fillDataSource = true
    }

    restore() {
      if (this._fillDataSource && this._mock !== undefined)
        throw new Error(
          `[DataSourceMotherObject#restore] consume DataSource before restore it`
        )
      DataSource.prototype.data = this._originalData
    }

    _response() {
      const res = {...this._mock}
      this._mock = undefined
      return res
    }
  }

export const RequestMotherObjectFactory = FetchFetcher =>
  class RequestMotherObject extends MotherObject {
    _url
    _method
    _error
    _mock

    constructor() {
      super()
      this._fetchFetcherFakerManager =
        FetchFetcherFakerManager.create(FetchFetcher)
      this._faker = FakeGenerator.create()
    }

    request({url, method = 'get'}) {
      this._url = url
      this._method = method
      return this
    }

    success() {
      if (this._error !== undefined)
        throw new Error(
          '[MotherObject#success] Dont call to .success after call to .error method'
        )

      this._mock = this.generate(this._faker)

      this._fetchFetcherFakerManager.addMock({
        url: this._url,
        method: this._method,
        mock: this._mock
      })
      return this
    }

    error({error}) {
      if (!error)
        throw new Error(
          '[MotherObject#error] if want to return an error from the API you must to instanciate a error'
        )

      if (this._mock !== undefined)
        throw new Error(
          '[MotherObject#error] Dont call to .error after call to .success method'
        )

      this._error = error
      this._fetchFetcherFakerManager.addMock({
        url: this._url,
        method: this._method,
        error: this._error
      })
      return this
    }

    toJSON() {
      return this._error ? this._error : this._mock
    }

    updateMock() {
      this._fetchFetcherFakerManager.addMock({
        url: this._url,
        method: this._method,
        mock: this._mock,
        force: true
      })
      return this
    }

    validate() {
      if (!this._url && !this._method) return
      this._fetchFetcherFakerManager.validate({
        url: this._url,
        method: this._method
      })
    }

    restore() {
      this._fetchFetcherFakerManager.restore()
    }
  }
